//#include <config.h>

#include <string.h>
#include <errno.h>
#include <glob.h>

#define DISABLE_PRINTING
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdkx.h>
extern "C" {
#include <gtkhtml/gtkhtml.h>
}
#include <hildon/hildon-banner.h>
#include <hildon/hildon-seekbar.h>
#include <hildon/hildon-touch-selector.h>
#include <hildon/hildon-picker-dialog.h>
//#include <gdk/gdk.h>
//#include <pango/pango-font.h>

#include "kmplayerplaylist.h"
#include "actor.h"
#include "kmplayer_lists.h"
#include "kmplayer_smil.h"
#include "kmplayercontrol.h"
#include "kmplayerprocess.h"
#include "kmplayer.h"
#include "viewarea.h"

using namespace KMPlayer;

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#ifdef KMPLAYER_DEBUG
static Application *debug_app;
static Control *debug_control;

void sig_usr1 (int) {
    debugLog () << "Start dump" << endl;
    debugLog () << debug_control->process () << " " <<
        debug_control->blanking_timer << endl;
    if (debug_control->process () && debug_control->process ()->mrl ())
        debugLog () << debug_control->process ()->playing () <<
            debug_control->process ()->mrl ()->src <<
            debug_control->process ()->mrl ()->audio_only <<
            debug_control->process ()->hasVideo () << endl;
    debugLog () << "End dump" << endl;
}
#endif

static gboolean playSchedule (void * ctr) {
#ifdef DEBUG_TIMERS
    debugLog() << "playSchedule" << endl;
#endif
    static_cast <Control *> (ctr)->play ();
    return false;
}

static gboolean playCurrentSchedule (void * ctr) {
#ifdef DEBUG_TIMERS
    debugLog() << "playCurrentSchedule" << endl;
#endif
    static_cast <Control *> (ctr)->playCurrent ();
    return false;
}

static Node *playlistMetaSelector (GtkWidget *selector, Node *meta, int index) {
    for (Node *c = meta->firstChild (); c; c = c->nextSibling ()) {
        if (id_node_meta_views == c->id) {
            for (Node *v = c->firstChild (); v; v = v->nextSibling ())
                if (id_node_meta_views == c->id) {
                    PlaylistRole *pr = (PlaylistRole *) v->role (RoleCaption);
                    if (pr) {
                        if (selector)
                            hildon_touch_selector_append_text (
                                    HILDON_TOUCH_SELECTOR (selector),
                                    pr->caption ());
                        else if (--index < 0)
                            return v;
                    }
                }
        } else if (id_node_playlist_item == c->id) {
            PlaylistRole *pr = (PlaylistRole *) c->role (RoleCaption);
            if (pr) {
                if (selector)
                    hildon_touch_selector_append_text (
                            HILDON_TOUCH_SELECTOR (selector),
                            pr->caption ());
                else if (--index < 0)
                    return c;
            }
        }
    }
    return NULL;
}

static void
playlistMetaResponse (GtkDialog *dialog, gint response, gpointer data) {
    Node **node = (Node **)data;
    HildonTouchSelector *sel = hildon_picker_dialog_get_selector (
            HILDON_PICKER_DIALOG (dialog));

    if (response == GTK_RESPONSE_OK) {
        int i = hildon_touch_selector_get_active(HILDON_TOUCH_SELECTOR(sel), 0);
        *node = playlistMetaSelector (NULL, *node, i);
    } else {
        *node = NULL;
    }
}

static void playlistMetaActivated (Node *meta, Application *app) {
    Control *ctrl = CONTROL(app);
    GtkWidget *dialog = hildon_picker_dialog_new (app->window);
    GtkWidget *selector = hildon_touch_selector_new_text();
    Node *selected_node = meta;

    hildon_picker_dialog_set_selector (HILDON_PICKER_DIALOG (dialog),
            HILDON_TOUCH_SELECTOR (selector));
    g_signal_connect (G_OBJECT (dialog), "response",
            G_CALLBACK (playlistMetaResponse), &selected_node);

    playlistMetaSelector (selector, meta, 0);

    PlaylistRole *pr = (PlaylistRole *) meta->role (RolePlaylist);
    if (pr)
        gtk_window_set_title (GTK_WINDOW (dialog), pr->caption ());

    gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_destroy (dialog);

    if (selected_node)
        jump (selected_node, ctrl);
}

SourceDocument::SourceDocument (Control *c, const String &url)
    : Document (url, c), m_control (c) {}

void SourceDocument::message (MessageType msg, void *data) {
    switch (msg) {

    case MsgInfoString:
        app_set_info_content (m_control->m_app,
                data ? (const char *) *((String *) data) : "", false);
        return;

    case MsgAccessKey:
        for (Connection *c = m_KeyListeners.first(); c; c = m_KeyListeners.next ())
            if (c->payload && c->connecter) {
                KeyLoad *load = (KeyLoad *) c->payload;
                if (load->key == (int) (long) data)
                    post (new Posting (this, c->connecter, MsgAccessKey));
            }
        return;

    case MsgListUpdated:
        for (PlayListData *pd = m_control->playlist_data; pd; pd = pd->next)
            if (pd->document.ptr () == this) {
                updateTree (m_control->m_app, pd->id);
                break;
            }
        return;

    case MsgShowMeta:
        playlistMetaActivated ((Node *) data, m_control->m_app);
        return;

    default:
        break;
    }
    Document::message (msg, data);
}

void *SourceDocument::role (RoleType msg, void *data) {
    switch (msg) {

    case RoleActorAgent:
        return m_control->actorAgent ();

    case RoleChildDisplay:
        return VIEWAREA(m_control->m_app)->getSurface ((Mrl *) data);

    case RoleReceivers:
        switch ((MessageType) (long) data) {

        case MsgAccessKey:
            return &m_KeyListeners;

        case MsgSurfaceUpdate:
            return VIEWAREA(m_control->m_app)->updaters ();

        default:
            break;
        }
        // fall through

    default:
        break;
    }
    return Document::role (msg, data);
}


Control::Control (Application * a)
    : m_process (0L), m_doc_timer (0), progress_banner (NULL), mplayer (NULL),
      playlist_data (0L), m_config (new Config), m_app (a),
      paint_timer (0), update_tree_timer (0), blanking_timer (0),
      in_progress_update (false),
      request_download (false),
      actor_agent (new ActorAgent (this)) {
#ifdef KMPLAYER_DEBUG
    debug_app = a;
    debug_control = this;
#endif
}

Control::~Control () {
    delete m_process;
    if (playlist_data->document)
        playlist_data->document->document()->dispose ();
    delete playlist_data;
    delete m_config;
    m_process = 0L;
}

static gboolean syncSchedule (Control *ctr)
{
    Recents *recents = static_cast <Recents *> (ctr->recents.ptr());
    if (!ctr->data_dir.isEmpty ()) {
        Playlist *playlist = static_cast<Playlist *>(ctr->playlists.ptr());
        FileDocument *downloads = (FileDocument*)ctr->downloads_doc.ptr ();
        if (recents)
            recents->sync (ctr->data_dir + "/recent.xml");
        if (downloads)
            downloads->sync (ctr->data_dir + "/downloads.xml");
        if (playlist)
            playlist->sync (ctr->data_dir + "/playlist.xml");
    }
    return !!recents;
}

void Control::init () {
    unsigned int ver, rel, req, evb, err;
    struct stat st;
    glob_t globbuf;

    use_xvideo = true;
   //media_server = new OssoMediaServer (this, m_app->osso_context, use_xvideo);
    MPlayer *gst = new MPlayer (this, "osso-media-server");
    gst->setPlayer ("/opt/kmplayer/bin/kgstplayer");
    gst->setAudioDriver ("pulse");
    gst->setVideoDriver ("xv");
    gst->setAutoSubTitles (true);
    media_server = gst;
    if (glob ("/usr/bin/mplayer", 0, NULL, &globbuf) != GLOB_NOMATCH) {
        MPlayer *mp = new MPlayer (this, "mplayer");
        mp->setAudioDriver ("pulse");
        mp->setVideoDriver ("xv:ck-method=auto");
        //mp->setAudioDriver ("pulse");
        mplayer = mp;
    }
    globfree (&globbuf);
#ifdef __ARMEL__
    npp = new NpPlayer (this);
    ((NpPlayer *)npp)->setPlayer ("/opt/kmplayer/bin/knpplayer");
#endif
    downloader = new Downloader (this, m_app);
    String p = m_config->readEntry (GCONF_KEY_PLAYER, "osso-media-server");
    setProcess ((const char *) p);
    selected_player = p;

    /*if (use_xvideo)
        use_xvideo = !m_config->readNumEntry (GCONF_KEY_VO, 0);
    if (!use_xvideo)
        gtk_check_menu_item_set_active (
                GTK_CHECK_MENU_ITEM (m_app->menu_item_mplayer_vo), true);*/

    GtkIconTheme *icon_theme = gtk_icon_theme_get_default();
    addTree (m_app, 0L, //gdk_pixbuf_new_from_xpm_data((const char**)www_xpm));
            gdk_pixbuf_scale_simple (
                    gtk_icon_theme_load_icon (icon_theme,
                        "general_web", 36,
                        (GtkIconLookupFlags)0, 0L),
                    36, 36, GDK_INTERP_BILINEAR));

    const String home = getenv ("HOME");
    if (!home.isEmpty ()) {
        data_dir = m_config->readEntry (GCONF_KEY_DATA_DIR,
                String (home + "/.kmplayer"));
        if (stat ((const char *) data_dir, &st))
            mkdir ((const char *) data_dir, 0755);
    }

    Playlist *pldoc = new Playlist (m_app, 0L);
    playlists = pldoc;
    String path = data_dir + "/playlist.xml";
    const char *tmp = (const char *) path;
    if (stat (tmp, &st)) {
        path = m_config->readEntry (GCONF_KEY_PLAYLIST_FILE,
                "/home/user/playlist.xml");
        tmp = (const char *) path;
        if (stat (tmp, &st))
            tmp = "/usr/share/applications/kmplayer/playlist.xml";
    }
    pldoc->path = tmp;

    m_app->playlists_tree_id = addTree (m_app, pldoc,
            gtk_icon_theme_load_icon(icon_theme,
                "kmplayer-fav", 36, (GtkIconLookupFlags)0, 0L));
    GConfClient *client = (GConfClient *) *m_config;
    GSList *plists = gconf_client_all_entries (client, GCONF_KEY_LISTS, NULL );
    for (GSList *elm = plists; elm; elm = elm->next) {
        GConfEntry *entry = (GConfEntry *)elm->data;
        GConfValue *value = gconf_entry_get_value (entry);
        const char *str = gconf_value_get_string (value);
        if (str && *str) {
            Playlist *expl = new Playlist (m_app, this);
            extra_playlists.append (expl);
            expl->title = entry->key + strlen (GCONF_KEY_LISTS) + 1;
            expl->path = str;
            addTree (m_app, expl,
                    gtk_icon_theme_load_icon(icon_theme,
                      "filemanager_playlist", 36, (GtkIconLookupFlags)0, 0L));
        }
        gconf_entry_free (entry);
    }
    g_slist_free (plists);

    Recents *rcdoc = new Recents (m_app);
    recents = rcdoc;
    path = data_dir + "/recent.xml";
    tmp = (const char *)path;
    if (stat (tmp, &st)) {
        String old = m_config->readEntry (GCONF_KEY_RECENT_FILE,
                "/home/user/recent.xml");
        if (!stat ((const char *) old, &st)) {
            path = old;
            tmp = (const char *) path;
        }
    }
    rcdoc->path = tmp;
    m_app->recents_tree_id = addTree (m_app, rcdoc,
            gtk_icon_theme_load_icon(icon_theme,
                "general_clock", 36, (GtkIconLookupFlags)0, 0L));

    String sf = m_config->readEntry (GCONF_KEY_SAVE_FOLDER, String());
    if (!sf.isEmpty ())
        m_app->save_folder = g_strdup ((const char *)sf);
    g_timeout_add (5*60*1000, GSourceFunc (syncSchedule), this);
}

void Control::deinit () {
    stop ();
    setProcess ((Process *) NULL);

    syncSchedule (this);

    Recents *rec = static_cast <Recents *> (recents.ptr());
    rec->dispose ();

    Playlist *playlist = static_cast<Playlist *>(playlists.ptr());
    if (m_app->save_folder)
        m_config->writeEntry (GCONF_KEY_SAVE_FOLDER, m_app->save_folder);
    playlist->dispose ();

    for (NodePtr n = extra_playlists.first(); n; n = n->nextSibling ())
        n->document()->dispose ();

    for (NodeStoreItem *s = generators.first (); s; s = generators.first ()) {
        s->data->document ()->dispose ();
        generators.remove (s);
    }

    if (downloads_doc) {
        FileDocument *doc = (FileDocument*)downloads_doc.ptr ();
        doc->dispose ();
        downloads_doc = NULL;
    }

    viewArea ()->deleteEvent ();

    delete media_server;
    delete mplayer;
#ifdef __ARMEL__
    delete npp;
#endif
    delete downloader;
}

static bool needsScreenBlanking (Application *app) {
    return app->top_most_window &&
        CONTROL(app)->playlist_data->current &&
        CONTROL(app)->playlist_data->document->state != Node::state_deferred &&
        (CONTROL(app)->playlist_data->current->id == SMIL::id_node_smil ||
         (CONTROL(app)->process ()->playing () &&
          CONTROL(app)->process ()->hasVideo ()));
}

gboolean cb_blankingTimer (void *p) {
#ifdef DEBUG_TIMERS
    debugLog() << "cb_blankingTimer" << endl;
#endif
    Application *app = static_cast <Application *> (p);
    if (needsScreenBlanking (app)) {
        osso_display_blanking_pause (app->osso_context);
        return 1;
    }
    CONTROL(app)->blanking_timer = 0;
    return 0;
}

gboolean cb_docTimer (void * p) {
#ifdef DEBUG_TIMERS
    debugLog() << "cb_docTimer" << endl;
#endif
    return static_cast <Control *> (p)->docTimerEvent ();
}

gboolean cb_scheduledUpdateTree (void * vp) {
#ifdef DEBUG_TIMERS
    debugLog() << "cb_scheduledUpdateTree" << endl;
#endif
    Control *ctr = static_cast <Control *> (vp);
    if (ctr->process ()) {
        if (!ctr->playlist_data->document)
            ctr->playlist_data->document = new SourceDocument (ctr, String());
        updateTree (ctr->m_app, 0);
    }
    ctr->update_tree_timer = 0;
    return false;
}

void Control::play () {
    if (!playlist_data->document)                // nothing to play
        return;
    if (playlist_data->document->active ()) {    // is playing
        Mrl *cur_item_playing = m_process->mrl ();
        if (!m_process->playing () ||
                Mrl::WindowMode == m_process->mrl ()->view_mode ||
                cur_item_playing == playlist_data->document->mrl ())
            cur_item_playing = NULL;

        if (playlist_data->document->state == Node::state_deferred) {
            playlist_data->document->undefer (); // unpause
            if (cur_item_playing)
                cur_item_playing->undefer ();
        } else {
            playlist_data->document->defer ();   // pause
            if (cur_item_playing)
                cur_item_playing->defer ();
        }
        app_update_screen_blanking (m_app);
    } else {
        if (!applyBackRequest () && playlist_data->document) {
            playCurrent ();
        }
    }
}

static PlayListData *playListData (PlayListData *pld, short id)
{
    while (pld->id != id)
        pld = pld->next;
    return pld;
}

void Control::stop (bool keep_fullscreen) {
    back_request = 0L;
    if (playlist_data->document &&
            playlist_data->document->state > Node::state_init) {
        playlist_data->document->reset ();
        setCurrent (0L);
    }
    VIEWAREA(m_app)->reset ();
    app_set_playing (m_app, NOT_PLAYING);

    if (current_generator && current_generator->active ()) {
        current_generator->deactivate ();
        current_generator = NULL;
    }

    Node *recent = recents->firstChild ();
    Mrl *mrl = m_process->mrl ();
    if (recent && mrl &&
            m_process->isLocalFile () &&
            recent->mrl ()->src == mrl->src) {
        Mrl *recent_mrl = recent->mrl ();
        String ts = recent_mrl->getAttribute ("clipBegin");
        int pos = m_process->userStopPosition ();
        pos = pos > 50 ? pos - 50 : 0;
        if (mrl->id == id_node_document)
            mrl->clip_begin = 10 * pos;
        String nts;
        int ds = pos % 10;
        if (ds > 0)
            nts = String (Char ('.')) + String::number (ds);
        pos /= 10;
        int s = pos % 60;
        nts = String::number (s) + nts;
        pos /= 60;
        if (pos > 0) {
            int m = pos % 60;
            nts = String::number (m) + String (Char (':')) + nts;
            pos /= 60;
            if (pos > 0) {
                if (m  < 10)
                    nts = String (Char ('0')) + nts;
                nts = String::number (pos) + String (Char (':')) + nts;
            }
        }
        if (ts != nts) {
            recent_mrl->setAttribute ("clipBegin", nts);
            playListData (playlist_data, m_app->recents_tree_id)
                ->updated_document_version =
                    ++recents->document ()->m_tree_version;
        }
    }
    m_process->quit ();
    setLoading (100); // no pogress bar

    VIEWAREA(m_app)->scheduleRepaint (IRect (0, 0,
                    VIEWAREA(m_app)->width (), VIEWAREA(m_app)->height ()));
    if (!keep_fullscreen)
        app_normal_screen (m_app);
    app_update_screen_blanking (m_app);
    setLength (0);
    setPosition (0);
    updatePlaylistView ();
}

String Control::currentMrl () {
    Mrl * mrl = playlist_data->current ? playlist_data->current->mrl () : 0L;
    if (!mrl)
        return String ();
    return mrl->absolutePath ();
}

void Control::openDocument (NodePtr doc, NodePtr cur) {
    stop (true);
    if (playlist_data->document)
        playlist_data->document->document()->dispose ();
    doc->document()->notify_listener = this;
    playlist_data->document = doc;
    playlist_data->current = cur ? cur : doc;
    cur_url = doc->document ()->caption ();
    app_set_title ((const char *) cur_url, m_app);
    back_request = cur;
    g_timeout_add (0, playSchedule, static_cast <void *> (this));
    updatePlaylistView ();
}

ViewArea *Control::viewArea () const {
    return (ViewArea *)m_app->viewarea;
}

void Control::playCurrent () {
    String url = currentMrl ();
    // set title url;
    aspect = 0.0;
    width = height = 0;
    //debugLog () << "playCurrent" << !!playlist_data->document << !playlist_data->document->active () << endl;
    // view ())->videoStop (); //show buttonbar
    if (playlist_data->document &&
            !playlist_data->document->active ()) {
        if (!playlist_data->current)
            setCurrent (playlist_data->document);
        else { // ugly code duplicate w/ back_request
            for (Node *p = playlist_data->current->parentNode(); p; p = p->parentNode())
                p->setState (Element::state_activated);
        }
        playlist_data->current->activate ();
        updateViewAreaSizes ();
    } else if (!playlist_data->current) {
        //debugLog () << "playCurrent no current" << endl;
        stop ();
        ; //emit endOfPlayItems ();
    } else if (playlist_data->current->state ==Element::state_deferred) {
        //debugLog () << "playCurrent set undefer" << endl;
        ;//playlist_data->current->undefer ();
    } else if (m_process->state () == Process::NotRunning) {
        //debugLog () << "playCurrent set ready" << endl;
        m_process->ready (VIEWAREA (m_app)->videoXId ());
    } else if (!playlist_data->current->active ()) {
        playlist_data->current->activate ();
        updateViewAreaSizes ();
    } else {
        Mrl *mrl = back_request
            ? back_request->mrl ()
            : playlist_data->current->mrl ();
        //debugLog () << "playCurrent " << m_process->playing () << " " << mrl->mrl ()->resolved << endl;
        if (m_process->playing ()) {
            errorLog() << "playCurrent and backend still playing" << endl;
            if (m_process->mrl () != mrl)
                m_process->stop ();
            else
                return; // FIXME
        }
        back_request = 0L;
        if (!mrl->src.isEmpty () && mrl->resolved) {
            updateViewAreaSizes ();
            if (m_process->hasXv ())
                XSync (GDK_DISPLAY_XDISPLAY (gtk_widget_get_display (VIEWAREA(m_app)->viewWidget ())), false);
            m_process->play (mrl);
        } else
            errorLog() << "playCurrent and mrl not resolved" << endl;
    }
}

bool Control::applyBackRequest() {
    if (!back_request || back_request->mrl ()->view_mode == Mrl::WindowMode)
        return false;
    if (playlist_data->current)
        playlist_data->document->reset ();
    playlist_data->current = 0L;
    if (back_request->document() != playlist_data->document) {
        //debugLog () << "playlist/recent" << endl;
        NodePtr n = back_request;
        back_request = 0L;
        n->activate();
    } else {
        setCurrent (back_request);
        for (Node *p = back_request->parentNode(); p; p = p->parentNode())
            p->setState (Element::state_activated);
        back_request = 0L;
        g_timeout_add(0, playCurrentSchedule, static_cast<void*>(this));
    }
    return true;
}

static String processForMimetype (Control *ctr, const String & mime)
{
    String pname = ctr->m_config->readEntry(String (GCONF_KEY_MIME) + mime, "");
#ifdef __ARMEL__
    if (pname.isEmpty ()) {
        // hardcoded defaults FIXME
        if (NpPlayer::isFlashMimeType (mime))
            pname = "npp";
    }
#endif
    return pname;
}

ActorAgent *Control::actorAgent () const {
    return actor_agent;
}

/* dbus-send --print-reply /com/nokia/osso_browser \
 * --dest=com.nokia.osso_browser com.nokia.osso_browser.load_url \
 *  string:http://maemo.org/
 */
bool Control::requestPlayURL (Mrl *mrl) {
    debugLog () << "requestPlayURL " << mrl->src << (mrl->view_mode == Mrl::SingleMode) << endl;
    if (process ()->state () > Process::Ready) {
        if (process ()->mrl () == mrl)
            return true;
        back_request = mrl; // still playing, schedule it
        if (!setProcess (mrl))
            process ()->stop ();
    } else {
        if (mrl->view_mode == Mrl::SingleMode)
            setCurrent (mrl);
        else
            back_request = mrl;
        setProcess (mrl);
        g_timeout_add (0, playCurrentSchedule, static_cast <void *> (this));
    }
    return true;
}

bool Control::setProcess (const char *pname) {
    Process *old_process = m_process;
    if (mplayer && !strcmp (pname, mplayer->name ())) {
        m_process = mplayer;
#ifdef __ARMEL__
    } else if (!strcmp (pname, npp->name ())) {
        m_process = npp;
#endif
    } else {
        m_process = media_server;
    }
    selected_player = m_process->name ();

    if (old_process) {
        if (old_process == m_process)
            old_process->stop ();
        else
            old_process->quit ();
    }
}

bool Control::setProcess (Mrl *mrl) {
    if (request_download) {
        m_process = downloader;
        request_download = false;
        return true;
    }
    String pname;
    bool changed = false;
    if (id_node_playlist_item == mrl->id) {
        String p = mrl->getAttribute ("player");
        if ((mplayer && p == mplayer->name ()) || p == media_server->name ()
#ifdef __ARMEL__
                || p == npp->name ()
#endif
           )
            pname = p;
    }
    if (pname.isEmpty () && !mrl->mimetype.isEmpty ())
        pname = processForMimetype (this, mrl->mimetype);
    if (pname.isEmpty ())
        pname = m_config->readEntry (GCONF_KEY_PLAYER, "osso-media-server");
    if (!m_process || pname != m_process->name ()) {
        setProcess ((const char *) pname);
        changed = true;
    }
    return changed;
}

void Control::setTimeout (int ms) {
    if (m_doc_timer > 0 && !g_source_remove (m_doc_timer))
        errorLog () << "failed to kill doc timer " << m_doc_timer << endl;

    m_doc_timer = ms > -1
        ? g_timeout_add (ms, cb_docTimer, static_cast <void *> (this))
        : 0;
}

void Control::openUrl (const URL &url, const String &, const String & /*srv*/) {
    osso_rpc_run_with_defaults( m_app->osso_context, "osso_browser",
            "open_new_window", NULL,
            DBUS_TYPE_STRING, (const char *)url.url (), DBUS_TYPE_INVALID);
    //errorMsg ("Open for target and mimetype is not supported");
}

void Control::enableRepaintUpdaters (bool enable, unsigned int off_time) {
    viewArea ()->enableUpdaters (enable, off_time);
}

bool Control::docTimerEvent () {
    int id = m_doc_timer;

    m_doc_timer = -1;

    if (playlist_data->document && playlist_data->document->active ())
        playlist_data->document->document ()->timer (); //can call setTimeout()
    else
        m_doc_timer = 0;
    if (-1 == m_doc_timer) {
        // not called setTimeout(), keep interval
        m_doc_timer = id;
        return true;
    }
    return false;
}

bool Control::setCurrent (NodePtr n) {
    debugLog() << "setCurrent " << (n && n->mrl() ? n->mrl()->src : "-")<< endl;
    if (playlist_data->current == n)
        return true;
    playlist_data->current = n;
    current_selected = n;
    updatePlaylistView ();
    Mrl *mrl = n ? n->mrl () : 0L;
    if (mrl) {
        String url;
        while (mrl) {
            if (mrl->src.startsWith ("Playlist://"))
                break;
            if (mrl->view_mode == Mrl::SingleMode)
                url = mrl->absolutePath ();
            Node *p = mrl->parentNode ();
            for (mrl = 0L; p; p = p->parentNode ())
                if (p->mrl ()) {
                    mrl = p->mrl ();
                    break;
                }
        }
        app_insert_location (m_app, (const char *)url);
    }
    app_update_screen_blanking (m_app);
    return true;
}

static const char *stateDesc [] = {
    "init", "deferred", "activated", "began", "finished",
    "deactivated", "resetting"
};

void Control::stateElementChanged (Node* n, Node::State os, Node::State ns) {
    //debugLog () << "state " << n->nodeName () << " changed to " << stateDesc [int (ns)] << endl;
    if (!m_process) // destructing
        return;
    if (ns == Node::state_deactivated &&
            n == playlist_data->document && !back_request) {
        stop ();
        //VIEWAREA(m_app)->setEventFiltering (m_app, false);
        app_set_playing (m_app, NOT_PLAYING);
    } else if ((ns == Node::state_deactivated || ns == Node::state_finished) &&
            n->mrl () && process ()->mrl () == n->mrl()) {
        process ()->stop ();
        updateViewAreaSizes (); // move away video widget
    } else if ((ns == Node::state_deferred ||
                (os == Node::state_deferred && ns > Node::state_deferred)) &&
            n == playlist_data->document) {
        //VIEWAREA(m_app)->setEventFiltering (m_app, true);
        app_set_playing (m_app, ns > Node::state_deferred ? PLAYING : PAUSED);
    } else if (ns == Node::state_activated ||
            (os == Node::state_init && ns == Node::state_deferred)) {
        if (n == playlist_data->document) {
            //VIEWAREA(m_app)->setEventFiltering (m_app, true);
            app_set_playing (m_app, PLAYING);
        }
        if (n->id == SMIL::id_node_body) {
            m_app->show_playlist = false;
            videoDetected ();
            updateViewAreaSizes (); // move away video widget
        }
        if (n->isPlayable () &&
                n->mrl ()->view_mode == Mrl::SingleMode) {
            Node *p = n->parentNode();
            if (!p || !p->mrl () || p->mrl ()->view_mode == Mrl::SingleMode)
                // make sure we don't set current to nested document
                setCurrent (n);
        }
    }
    if (n == playlist_data->current &&
            (ns == Node::state_activated || ns == Node::state_deactivated))
        updatePlaylistView ();
}

void Control::updateViewAreaSizes () {
    if (!m_process || !VIEWAREA(m_app)->videoXId ())
        return;
    VIEWAREA(m_app)->resizeEvent();
    IRect video_rect = VIEWAREA(m_app)->videoGeometry ();
    int x = video_rect.x ();
    int y = video_rect.y ();
    m_process->setVideoWindow (x, y, video_rect.width (), video_rect.height ());
}

void Control::updatePlaylistView (bool force) {
    if (force) {
        if (update_tree_timer) {
            g_source_remove (update_tree_timer);
            update_tree_timer = 0;
        }
        updateTree (m_app, 0);
    } else if (!update_tree_timer) {
        update_tree_timer = g_timeout_add (200, cb_scheduledUpdateTree, this);
    }
}

static const char * statemap [] = {
        "Not Running", "Ready", "Buffering", "Playing", "Paused"
};

void Control::stateChanged (Process * p, Process::State os, Process::State ns) {
    Mrl *mrl = p->mrl ();
    NodePtrW guard = mrl; // p->mrl is weak, needs check
    debugLog () << "Process stateChanged " << statemap [(int)os] << " -> " << statemap [(int)ns] << endl;
    if (!m_process) // destructing
        return;
    if (!mrl && ns > Process::Ready) {
        p->stop (); // reschedule to the NotRunning state
    } else if (ns == Process::NotRunning) {
        if (p == m_process) {
            if (playlist_data->document &&
                    playlist_data->document->state > Node::state_init) {
                stop (); // this should not happen
            }
            app_update_screen_blanking (m_app);
        }
        // else process changed
        if (!playlist_data->edit_mode)
            app_set_info_content (m_app, "", false);
    } else if (ns == Process::Ready) {
        setLoading (100); // move possible buffer info dialog
        if (os > Process::Ready) {
            setLength (0);
            setPosition (0);
            VIEWAREA(m_app)->setAspect (0.0);
            if (guard && !applyBackRequest () &&  // if cause is no jump
                    mrl->active ()) {           // if cause is eof
                mrl->message (MsgMediaFinished);// set node to finished
                if (back_request)            // overlapping audio/video in SMIL
                    g_timeout_add(0, playCurrentSchedule, (void *) this);
            }
        } else
            g_timeout_add (0, playCurrentSchedule, static_cast <void *> (this));
    } else if (ns == Process::Buffering) {
        if (mrl->view_mode != Mrl::SingleMode) {
            mrl->defer ();
        }
    } else if (ns == Process::Paused &&
            playlist_data->document->state != Node::state_deferred &&
            !playlist_data->document->document ()->postponed ()) {
        if (gtk_window_has_toplevel_focus (m_app->window))
            m_process->pause ();
        else
            playlist_data->document->defer ();
    } else if (ns == Process::Playing) {
        mrl->audio_only |= !p->hasVideo ();
        if (mrl->state == Element::state_deferred)
            mrl->undefer ();
        setLoading (100); // move possible buffer info dialog
        addRecent (playlist_data->document->mrl()->src,
                playlist_data->document->document ()->caption ());
        updatePlayTree (m_app);
        if (mrl->audio_only &&
                playlist_data->current &&
                playlist_data->current->id != SMIL::id_node_smil)
            app_normal_screen (m_app);
        app_update_screen_blanking (m_app);
    }
}

void Control::addRecent (const String &url, const String &title) {
    if (url == "Playlist://")
        return;
    recents->defer ();
    Node *c = recents->firstChild ();
    if (!c || c->mrl()->src != url) {
        Recent *rec = new Recent (recents, m_app, url);
        recents->insertBefore (rec, c);
        if (!title.isEmpty ())
            rec->setAttribute (Ids::attr_title, title);
        rec->closed ();
        c = recents->firstChild ()->nextSibling ();
        int count = 1;
        Node *more = NULL;
        while (c) {
            if (c->id == id_node_recent_node &&
                    c->mrl ()->src == url) {
                Node *tmp = c->nextSibling ();
                recents->removeChild (c);
                c = tmp;
            } else {
                if (c->id == id_node_group_node)
                    more = c;
                c = c->nextSibling ();
                count++;
            }
        }
        if (!more && count > 10) {
            more = new Group (recents, m_app, "More ...");
            recents->appendChild (more);
        }
        if (more) {
            if (count > 10) {
                NodePtr item = more->previousSibling ();
                recents->removeChild (item);
                more->insertBefore (item.ptr (), more->firstChild ());
            }
            if (more->firstChild ())
                c = more->firstChild ()->nextSibling ();
            count = 0;
            while (c) {
                if (c->id == id_node_recent_node &&
                        c->mrl ()->src == url) {
                    Node *tmp = c->nextSibling ();
                    more->removeChild (c);
                    c = tmp;
                } else {
                    c = c->nextSibling ();
                    count++;
                }
            }
            if (count > 50)
                more->removeChild (more->lastChild ());
        }
        playListData(playlist_data, m_app->recents_tree_id)->current = NULL;
        updateTree (m_app, m_app->recents_tree_id);
    }
}

void Control::setCurrentSelected (NodePtr node) {
    current_selected = node;
}

void Control::errorMsg (const String & msg) {
    hildon_banner_show_information (GTK_WIDGET (m_app->window),
            NULL, (const char *) msg);
}

void Control::setLoading (int perc) {
    if (perc >= 100) {
        if (progress_banner)
            gtk_widget_destroy (progress_banner);
        progress_banner = NULL;
    } else {
        if (!progress_banner)
            progress_banner = hildon_banner_show_progress (
                    GTK_WIDGET (m_app->window), NULL, "Buffering");
        hildon_banner_set_fraction (
                HILDON_BANNER (progress_banner), 1.0 * perc / 100);
    }
}

void Control::setPosition (int p) {
    if (p < hildon_seekbar_get_total_time(HILDON_SEEKBAR(m_app->progress_bar))){
        in_progress_update = true;
        if (progress_banner)
            setLoading (100);
        if (p > 0)
            hildon_seekbar_set_position(HILDON_SEEKBAR(m_app->progress_bar), p);
        in_progress_update = false;
    }
}

void Control::setLength (int len) {
    hildon_seekbar_set_total_time (HILDON_SEEKBAR (m_app->progress_bar), len);
    hildon_seekbar_set_fraction (HILDON_SEEKBAR (m_app->progress_bar), len);
    m_app->show_progress = len > 50;
    app_set_playing (m_app, m_app->playing);
}

void Control::setAspect (float aspect) {
    VIEWAREA(m_app)->setAspect (aspect);
    updateViewAreaSizes ();
}

void cb_view_area_configure (GtkWidget *widget, GdkEventConfigure *event, Application * app) {
    //debugLog () << "cb_view_area_configure " << event->width<< "x" << event->height << endl;
    CONTROL(app)->updateViewAreaSizes ();
}

void Control::bitRates (int & prefered, int & maximal) {
}

PlayListData::PlayListData(short i, PlayListData * n)
 : updated_document (NULL),
   updated_document_version (0),
   id (i),
   show_all (false),
   edit_mode (false),
   next (n) {}

Application *app_new () {
    Application *app = new Application;
    memset (app, 0, sizeof (Application));
    app->control = new Control (app);
    app->show_playlist = true;
    Ids::init();
    app->viewarea = new ViewArea (app);
    app->top_most_window = true;
    return app;
}

void app_destroy (Application *app) {
    delete CONTROL(app);
    delete VIEWAREA(app);
    Ids::reset ();
}

void app_init_control (Application *app) {
    CONTROL(app)->init ();
}

void app_update_screen_blanking (Application *a) {
    bool needs_blanking_pause = needsScreenBlanking (a);
    if (needs_blanking_pause && !CONTROL(a)->blanking_timer) {
        CONTROL(a)->blanking_timer = g_timeout_add (25000, cb_blankingTimer, a);
        osso_display_blanking_pause (a->osso_context); // 60s delay
    } else if (!needs_blanking_pause && CONTROL(a)->blanking_timer) {
        g_source_remove (CONTROL(a)->blanking_timer);
        CONTROL(a)->blanking_timer = 0;
    }
}

static Mrl *findFirstPlayable (Node *n) {
    if (n) {
        if (n->isPlayable ())
            return n->mrl ();
        for (Node *c = n->firstChild (); c; c = c->nextSibling ()) {
            Mrl *mrl = findFirstPlayable (c);
            if (mrl)
                return mrl;
        }
    }
    return NULL;
}

void KMPlayer::jump (Node *n, Control *ctr) {
    Mrl *mrl = n->document() != ctr->playlist_data->document
        ? isListItem (n->id) ||
          id_node_meta_view == n->id ||
          id_node_meta_group == n->id
              ? n->mrl ()
              : NULL
        : findFirstPlayable (n);
    if (mrl && mrl->view_mode == Mrl::SingleMode) {
        if (n->document() == ctr->playlist_data->document && !mrl->isPlayable())
            return;
        if (ctr->process ()->playing () &&
                ctr->process ()->mrl() == mrl) {
            ctr->play (); // pause or unpause
        } else {
            //debugLog() << "jump " << mrl->src << endl;
            // FIXME: group activation doesn't work
            ctr->back_request = mrl;
            if (ctr->process ()->playing ())
                ctr->process ()->stop ();
            else
                ctr->applyBackRequest();
        }
    } else
        ctr->updatePlaylistView ();
}

void app_set_info_content (Application *app, const char *s, bool plaintext) {
    if (!app->fullscreen) {
        if (!s || !*s)
            gtk_widget_hide (app->info_area);
        else
            gtk_widget_show (app->info_area);
    }
    if (!plaintext) {
        gtk_widget_hide (app->info_viewport);
        gtk_widget_show (app->html_viewport);
        if (!s[0]) {
            gtk_html_load_empty (GTK_HTML (app->html_infopanel_view));
        } else {
            int w, h;
            gtk_widget_get_size_request (app->info_area, &w, &h);
            if (h != 140) {
                gtk_widget_set_size_request (app->info_area, 400, 140);
                gtk_fixed_move (GTK_FIXED (app->window_area),
                        app->info_area, 400, 214);
            }
            gtk_html_load_from_string(GTK_HTML(app->html_infopanel_view), s,-1);
        }
    } else {
        int w, h;
        gtk_widget_get_size_request (app->info_area, &w, &h);
        if (h != 354) {
            gtk_fixed_move (GTK_FIXED (app->window_area),
                    app->info_area, 400, 0);
            gtk_widget_set_size_request (app->info_area, 400, 354);
        }
        gtk_widget_hide (app->html_viewport);
        gtk_text_buffer_set_text (gtk_text_view_get_buffer
                (GTK_TEXT_VIEW (app->infopanel_view)), s, -1);
        gtk_widget_show (app->info_viewport);
    }
}

void cbPluginAdded (GtkSocket *socket, Application *app) {
#ifdef __ARMEL__
    Control *ctr = CONTROL(app);
    debugLog () << "plugin embedded\n";
    app->show_playlist = false;
    ctr->videoDetected ();
#endif
}

void cb_playlist_prev (GtkButton * button, void *p) {
    Control *ctr = (Control *)p;
    if (ctr->playlist_data->current &&
            ctr->playlist_data->current->previousSibling ())
        jump (ctr->playlist_data->current->previousSibling (), ctr); //FIXME
}

void cb_playlist_play (GtkButton * button, void *p) {
    Control *ctr = (Control *)p;
    if (button == GTK_BUTTON (ctr->m_app->tb_pause) ||
            (ctr->playlist_data->document &&
             ctr->playlist_data->document->state == Node::state_deferred))
        ctr->play ();
    else if (ctr->current_selected)
        jump (ctr->current_selected, ctr);
    else
        g_timeout_add (0, playSchedule, static_cast <void *> (ctr));
}

void cb_playlist_next (GtkButton * button, void *p) {
    Control *ctr = (Control *)p;
    if (ctr->process ()->playing ())
        ctr->process ()->stop ();
}

void cb_playlist_stop (GtkButton * button, void *p) {
    Control *ctr = (Control *)p;
    if (GTK_WIDGET_VISIBLE (ctr->m_app->tb_stop)) {
        ctr->process ()->setUserStop ();
        ctr->stop ();
    } else if (ctr->current_selected) {
        ctr->request_download = true;
        jump (ctr->current_selected, ctr);
    } else {
        hildon_banner_show_information (GTK_WIDGET (ctr->m_app->window), NULL,
                "No Playlist Item Selected");
    }
}

void cb_favorites_add (GtkButton *, Application *app) {
    Control *ctr = CONTROL(app);
    bool valid = false;
    const gchar *msg = "Unknown failure";
    if (!ctr->current_selected) {
        msg = "No current selected";
    } else {
        Mrl *mrl = ctr->current_selected->mrl ();
        if (!mrl || mrl->src.isEmpty() || mrl->src.startsWith ("Playlist://")) {
            msg = "Not a valid link selected";
        } else {
            PlayListData *pd, *fav = NULL;
            for (pd = ctr->playlist_data; pd; pd = pd->next) {
                if (pd->id == app->playlists_tree_id)
                    fav = pd;
                if (pd->document.ptr () == mrl->document ()) {
                    if (fav == pd)
                        msg = "Already in favorites";
                    else
                        valid = true;
                }
            }
            if (fav && valid) {
                msg = "Added to favorites";
                NodePtr doc = mrl->document ();
                if (!fav->document->mrl ()->resolved)
                    fav->document->activate (); // lazy load playlist
                if (id_node_playlist_item == mrl->id ||
                        id_node_html_object == mrl->id) {
                    fav->document->insertBefore (mrl->outerXML (false), NULL);
                } else {
                    PlaylistItem *i = new PlaylistItem (doc, app, false, mrl->src);
                    if (!mrl->title.isEmpty ())
                        i->setAttribute (Ids::attr_title, mrl->title);
                    i->closed ();
                    fav->document->appendChild (i);
                }
                updateTree (app, fav->id);
            }
        }
    }
    hildon_banner_show_information (GTK_WIDGET (app->window), NULL, msg);
}

void openFile (Application * app, const gchar * path) {
    CONTROL(app)->stop ();
    PlayListData *pd = CONTROL(app)->playlist_data;
    if (pd->document)
        pd->document->document()->dispose ();
    pd->document = new SourceDocument (CONTROL(app), String (path));
    CONTROL(app)->cur_url = path;
    g_timeout_add (0, playSchedule, static_cast <void *> (CONTROL(app)));
    CONTROL(app)->updatePlaylistView (true);
}

void cb_openUrls (void *app, int argc, gchar **argv) {
}

static String makeM3u (Node *n, const String &group, const String &fn) {
    String str;
    if (id_node_playlist_item == n->id || id_node_recent_node == n->id) {
        Mrl *mrl = n->mrl ();
        if (!mrl->src.isEmpty () && fn != mrl->src) {
            if (!mrl->caption ().isEmpty ())
                str = String ("#EXTINF:,") +
                    (id_node_playlist_item == n->id && !group.isEmpty ()
                     ? group + " - "
                     : String ()) +
                    mrl->caption () + String ("\r\n");
            str += mrl->src + String ("\r\n");
        }
    } else {
        String title;
        if (id_node_group_node == n->id) {
            PlaylistRole *title_role = (PlaylistRole *) n->role (RolePlaylist);
            if (title_role)
                title = title_role->caption ();
        }
        for (Node *c = n->firstChild (); c; c = c->nextSibling ())
            str += makeM3u (c, title, fn);
    }
    return str;
}

void saveFile (Application *app, bool xml) {
    Control *ctr = CONTROL(app);
    String msg;
    if (ctr->current_selected) {
        Document *doc = ctr->current_selected->document ();
        gchar *filename = app_file_chooser (app, app->save_folder,
                GTK_FILE_CHOOSER_ACTION_SAVE);
        if (filename) {
            String fn (filename);
            String str;
            if (xml) {
                if (!fn.lower ().endsWith (".xml"))
                    fn += ".xml";
                str = doc->outerXML ();
            } else {
                if (!fn.lower ().endsWith (".m3u"))
                    fn += ".m3u";
                str = String ("#EXTM3U\r\n") + makeM3u (doc, String (), fn);
            }
            int fd = open ((const char*)fn, O_WRONLY | O_CREAT | O_TRUNC, 0664);
            if (fd < 0) {
                msg = String( "Failed creating ") + fn + Char ('\n') + strerror (errno);
            } else {
                const char *txt = (const char *) str;
                ::write (fd, txt, strlen (txt));
                msg = String ("Saved ") + fn;
                ::close (fd);
            }
            app_store_save_directory (app, filename); // free's filename
        }
    } else {
        msg = "No list selected";
    }
    if (!msg.isEmpty ())
        hildon_banner_show_information (GTK_WIDGET (app->window), NULL,
                (const char *)msg);
}

void newFile (Application * app) {
    PlayListData *pd = CONTROL(app)->playlist_data;
    if (pd->document) {
        pd->document->reset ();
        pd->document->document ()->dispose ();
    }
    pd->document = new SourceDocument (CONTROL(app), String ());
    CONTROL(app)->updatePlaylistView (true);
    CONTROL(app)->cur_url = String ();
}

gboolean cb_mainview_delete (GtkObject *, GdkEvent *, Application *app) {
    CONTROL(app)->deinit ();
    gtk_main_quit ();
    return false;
}

static bool testFullscreenPlaylist (Application *app) {
    Control *ctr = CONTROL(app);
    return ((ctr->process () != ctr->mplayer || !ctr->process ()->hasXv ()) &&
            app->fullscreen);
}

gboolean cb_keyPressed (GtkWidget *, GdkEventKey * e, Application *app) {
    switch (e->keyval) {
        case GDK_F6: //HILDON_HARDKEY_FULLSCREEN:
            //if (app->control->process ()->playing ()) {
             //   int w, h;
            VIEWAREA(app)->prepareFullscreenToggle ();
            if (!app->fullscreen) {
                if (CONTROL(app)->playlist_data->document &&
                        CONTROL(app)->playlist_data->document->active ())
                    app_full_screen (app);
            } else {
                app_normal_screen (app);
            }
            break;
        case GDK_F7: //HILDON_HARDKEY_INCREASE
            if (CONTROL(app)->process ()->playing ())
                CONTROL(app)->process ()->volume (2, false);
            break;
        case GDK_F8: //HILDON_HARDKEY_DECREASE
            if (CONTROL(app)->process ()->playing ())
                CONTROL(app)->process ()->volume (-2, false);
            break;
        case GDK_Left:
            if (gtk_window_get_focus (app->window) == app->playlist_view ||
                    testFullscreenPlaylist (app))
                collapseSelectedList (app);
            else if (CONTROL(app)->process ()->playing ())
                CONTROL(app)->process ()->seek (-200, false); // 20s
            break;
        case GDK_Right:
            if (gtk_window_get_focus (app->window) == app->playlist_view ||
                    testFullscreenPlaylist (app))
                expandSelectedList (app);
            else if (CONTROL(app)->process ()->playing ())
                CONTROL(app)->process ()->seek (200, false); // 20s
            break;
        case GDK_KP_Enter:
        case GDK_Return: // HILDON_HARDKEY_SELECT
            if (gtk_window_get_focus (app->window) == app->playlist_view ||
                    testFullscreenPlaylist (app))
                rowSelected (app);
            break;
        case GDK_Up:
            if (gtk_window_get_focus (app->window) == app->playlist_view ||
                    testFullscreenPlaylist (app))
                previousSelected (app);
            break;
        case GDK_Down:
            if (gtk_window_get_focus (app->window) == app->playlist_view ||
                    testFullscreenPlaylist (app))
                nextSelected (app);
            break;
        case GDK_Escape:
            return true;
        default:
            KMPlayer::debugLog () << "Keypress " << e->keyval << KMPlayer::endl;
    }
    return false;
}

void progressValueChanged(GtkWidget *w, Application * app) {
    if (!CONTROL(app)->in_progress_update &&
            CONTROL(app)->process ()->playing ()) {
        int p = hildon_seekbar_get_position(HILDON_SEEKBAR (app->progress_bar));
        CONTROL(app)->process ()->seek (p, true);
        //debugLog () << "progressBarClicked " << p << endl;
        app_show_controls (app);
    }
}

void app_set_suspend (Application *a, bool suspend) {
    PlayListData *pd = CONTROL(a)->playlist_data;
    if (pd->document &&
            pd->document->active () &&
            (!CONTROL(a)->process ()->playing () ||
             CONTROL(a)->process()->mrl()->view_mode ==Mrl::WindowMode ||
             CONTROL(a)->process ()->hasVideo ())) {
        if (pd->document->state == Node::state_deferred) {
            if (!suspend)
                ; // do nothing  a->playlist_data->document->undefer ();
        } else if (suspend)
            pd->document->defer ();
    }
}

static bool locationbar_initialized;

gboolean
locationbarGainFocus (GtkWidget *, GdkEventFocus *, Application * app) {
    int n = gtk_toolbar_get_n_items (GTK_TOOLBAR (app->main_toolbar));
    for (int i = 5; i < n; i++)
        gtk_widget_hide (GTK_WIDGET (gtk_toolbar_get_nth_item (GTK_TOOLBAR (app->main_toolbar), i)));
    app_set_playing (app, app->playing);
    if (!locationbar_initialized) {
        const char * defs[6] = {
            "http://www.cwi.nl/SMIL/EUROshow/EUROshowB.smi",
            "http://homepages.cwi.nl/~media/SMIL/fiets/fiets.smil",
            "http://www.theworld.org/rss/tech.xml",
            "http://www.etherbeat.com/shoutcast/64.pls",
            "http://www.swissgroove.ch/listen.m3u",
            "file:///home/user/MyDocs/.videos/Nokia_N900.mp4"
        };
        int defpos = 0;
        locationbar_initialized = true;
        KMPlayer::Config * config = CONTROL(app)->m_config;
        KMPlayer::String cfg_prefix (GCONF_KEY_LOCATION_URL);
        GtkTreeIter iterator;
        GtkListStore * model = GTK_LIST_STORE (gtk_combo_box_get_model (
                    GTK_COMBO_BOX (app->location_bar)));
        for (int i = 1; i < 7; i++) {
            gtk_list_store_append (model, &iterator);
            KMPlayer::String cfg = config->readEntry (cfg_prefix +
                    KMPlayer::String::number(i), "");
            if (cfg.isEmpty ())
                cfg = defs[defpos++];
            gtk_list_store_set (model, &iterator,
                0, (const char *) cfg, -1);
        }
    }
    gtk_editable_select_region (GTK_EDITABLE (GTK_BIN (app->location_bar)->child), 0, -1);
    //g_print ("callback_locationbar_gain_focus %d\n", n);
    return false;
}

void save_locationbar (Application *app) {
    if (!locationbar_initialized)
        return;
    KMPlayer::Config * config = CONTROL(app)->m_config;
    int idx = 1;
    KMPlayer::String cfg_pfx (GCONF_KEY_LOCATION_URL);
    GtkTreeIter iterator;
    GtkTreeModel * model = GTK_TREE_MODEL (gtk_combo_box_get_model (
                GTK_COMBO_BOX (app->location_bar)));
    for (bool b = gtk_tree_model_get_iter_first (model, &iterator);
            b;
            b = gtk_tree_model_iter_next (model, &iterator)) {
        GValue value = { 0 , };
        gtk_tree_model_get_value (model, &iterator, 0, &value);
        KMPlayer::String val (g_value_get_string (&value));
        g_value_unset (&value);
        if (!val.isEmpty ())
            config->writeEntry (cfg_pfx + KMPlayer::String::number(idx++), val);
    }
}

static void
playerResponse (GtkDialog *dialog, gint response, gpointer data) {
    if (response == GTK_RESPONSE_OK) {
        Control *ctrl = (Control *) data;
        HildonTouchSelector *sel = hildon_picker_dialog_get_selector (
                HILDON_PICKER_DIALOG (dialog));

        ctrl->selected_player = hildon_touch_selector_get_current_text (sel);
    }
}

void playerMenuActivate (GtkMenuItem *item, Application *app) {
    Control *ctrl = CONTROL(app);
    GtkWidget *dialog = hildon_picker_dialog_new (app->window);
    GtkWidget *selector = hildon_touch_selector_new_text();
    String player = ctrl->selected_player;
    int idx = 0;

    hildon_picker_dialog_set_selector (HILDON_PICKER_DIALOG (dialog),
            HILDON_TOUCH_SELECTOR (selector));
    g_signal_connect (G_OBJECT (dialog), "response",
            G_CALLBACK (playerResponse), ctrl);
    hildon_touch_selector_append_text (HILDON_TOUCH_SELECTOR (selector),
            ctrl->media_server->name ());
    if (ctrl->mplayer) {
        hildon_touch_selector_append_text (HILDON_TOUCH_SELECTOR (selector),
                ctrl->mplayer->name ());
        if (player == ctrl->mplayer->name ())
            idx = 1;
    }
#ifdef __ARMEL__
    hildon_touch_selector_append_text (HILDON_TOUCH_SELECTOR (selector),
            ctrl->npp->name ());
    if (player == ctrl->npp->name ())
        idx = ctrl->mplayer ? 2 : 1;
#endif
    hildon_touch_selector_set_active (HILDON_TOUCH_SELECTOR (selector), 0, idx);
    gtk_window_set_title (GTK_WINDOW (dialog), "Select Player");
    gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_destroy (dialog);

    if (player != ctrl->selected_player) {
        Process *p = ctrl->media_server;
        if (ctrl->mplayer && ctrl->selected_player == ctrl->mplayer->name ())
            p = ctrl->mplayer;
#ifdef __ARMEL__
        else if (ctrl->selected_player == ctrl->npp->name ())
            p = ctrl->npp;
#endif
        ctrl->setProcess (p->name ());
#ifdef __ARMEL__
        if (p != ctrl->npp)
#endif
            ctrl->m_config->writeEntry(GCONF_KEY_PLAYER, String(p->name()));
    }
}

static void getGenerators (Control *ctrl, const String &dir) {
    GDir *gdir = g_dir_open ((const char *) dir, 0, NULL);
    if (gdir) {
        const gchar *file = g_dir_read_name (gdir);
        while (file) {
            Generator *gen = new Generator (ctrl);
            NodePtr doc = gen;
            gen->readFromFile (dir + Char ('/') + file);
            Node *n = gen->firstChild ();
            if (n && n->isElementNode ()) {
                String txt = static_cast <Element *> (n)->getAttribute (
                        Ids::attr_name);
                if (txt.isEmpty ())
                    txt = file;
                gen->setAttribute (Ids::attr_name, txt);
                ctrl->generators.append (new NodeStoreItem (doc));
            } else {
                gen->dispose ();
            }
            file = g_dir_read_name (gdir);
        }
        g_dir_close (gdir);
    }
}

static void
generatorResponse (GtkDialog *dialog, gint response, gpointer data) {
    Control *ctrl = (Control *) data;
    HildonTouchSelector *sel = hildon_picker_dialog_get_selector (
            HILDON_PICKER_DIALOG (dialog));

    if (response == GTK_RESPONSE_OK) {
        gchar *selection = hildon_touch_selector_get_current_text (sel);
        for (NodeStoreItem *s = ctrl->generators.first(); s; s = s->nextSibling()) {
            String txt = static_cast <Element *> (s->data.ptr())->getAttribute (
                    Ids::attr_name);
            if (txt == selection) {
                ctrl->current_generator = s->data;
                return;
            }
        }
    }
    ctrl->current_generator = NULL;
}

void generatorMenuActivate (GtkMenuItem *item, Application *app) {
    Control *ctrl = CONTROL(app);
    GtkWidget *dialog = hildon_picker_dialog_new (app->window);
    GtkWidget *selector = hildon_touch_selector_new_text();

    hildon_picker_dialog_set_selector (HILDON_PICKER_DIALOG (dialog),
            HILDON_TOUCH_SELECTOR (selector));
    g_signal_connect (G_OBJECT (dialog), "response",
            G_CALLBACK (generatorResponse), ctrl);

    if (!ctrl->generators.first ()) {
        getGenerators (ctrl, "/usr/share/applications/kmplayer/generators");
        getGenerators (ctrl, ctrl->data_dir + "/generators");
    } else if (ctrl->current_generator) {
        if (ctrl->current_generator->active ())
            ctrl->current_generator->deactivate ();
        ctrl->current_generator = NULL;
    }

    for (NodeStoreItem *s = ctrl->generators.first(); s; s = s->nextSibling()) {
        String txt = static_cast <Element *> (s->data.ptr ())->getAttribute (
                Ids::attr_name);
        hildon_touch_selector_append_text (HILDON_TOUCH_SELECTOR (selector),
                                txt);
    }
    gtk_window_set_title (GTK_WINDOW (dialog), "Generate Playlist");
    gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_destroy (dialog);

    if (ctrl->current_generator)
        ctrl->current_generator->activate ();
}

static void
playlistResponse (GtkDialog *dialog, gint response, gpointer data) {
    if (response == GTK_RESPONSE_OK) {
        HildonTouchSelector *sel = hildon_picker_dialog_get_selector (
                HILDON_PICKER_DIALOG (dialog));
        *(int *) data = hildon_touch_selector_get_active (
                HILDON_TOUCH_SELECTOR (sel), 0);
    } else {
        *(int *) data = -1;
    }
}

void playlistMenuActivate (GtkMenuItem *, Application *app) {
    Control *ctrl = CONTROL(app);
    if (ctrl->current_selected) {
        GtkWidget *dialog = hildon_picker_dialog_new (app->window);
        GtkWidget *selector = hildon_touch_selector_new_text();
        int idx = 0;

        hildon_picker_dialog_set_selector (HILDON_PICKER_DIALOG (dialog),
                HILDON_TOUCH_SELECTOR (selector));
        g_signal_connect (G_OBJECT (dialog), "response",
                G_CALLBACK (playlistResponse), &idx);
        hildon_touch_selector_append_text (HILDON_TOUCH_SELECTOR (selector),
                "Export as XML");
        hildon_touch_selector_append_text (HILDON_TOUCH_SELECTOR (selector),
                "Export as m3u");
        hildon_touch_selector_append_text (HILDON_TOUCH_SELECTOR (selector),
                "Change Title");
        hildon_touch_selector_append_text (HILDON_TOUCH_SELECTOR (selector),
                "Show all XML Nodes");
        hildon_touch_selector_append_text (HILDON_TOUCH_SELECTOR (selector),
                "Edit XML");
        gtk_window_set_title (GTK_WINDOW (dialog), "Playlist Options");
        gtk_dialog_run (GTK_DIALOG (dialog));
        gtk_widget_destroy (dialog);

        switch (idx) {
        case 0:
            saveFile (app, true);
            break;
        case 1:
            saveFile (app, false);
            break;
        case 2:
            cb_playlistTitle (app);
            break;
        case 3:
            togglePlaylistShowAll (app);
            break;
        case 4:
            togglePlaylistEdit (app);
            break;
        }
    } else {
        hildon_banner_show_information (GTK_WIDGET (app->window), NULL,
                "No playlist item selected");
    }
}

#include <hildon/hildon-pannable-area.h>
#include <hildon/hildon-program.h>
#include <hildon/hildon-window.h>
#include <hildon/hildon-defines.h>
#include <hildon/hildon-gtk.h>

namespace {

class DownloadDocument : public FileDocument
{
public:
    DownloadDocument (Control *ctrl)
        : FileDocument (id_node_download_document, "", ctrl)
    {
        readFromFile (ctrl->data_dir + "/downloads.xml");
        resolved = true;
    }

    virtual Node *childFromTag (const String &tag)
    {
        if (tag == "download")
            return new Download (m_doc, m_control);
        if (tag == "downloads")
            return this;
        return FileDocument::childFromTag (tag);
    }
    virtual const char *nodeName () const {
        return "downloads";
    }
};

}

Node *Control::downloads ()
{
    if (!downloads_doc)
        downloads_doc = new DownloadDocument (this);
    return downloads_doc.ptr ();
}

void Control::updateDownloads ()
{
    if (m_app->downloads_view) {
        GtkListStore *store = GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (m_app->downloads_view)));
        gtk_list_store_clear (store);
        for (Node *c = downloads ()->firstChild (); c; c = c->nextSibling ())
            if (id_node_download == c->id) {
                GtkTreeIter iter;
                gtk_list_store_append (store, &iter);
                Download *download = static_cast <Download *> (c);
                gtk_list_store_set (store, &iter,
                        0, (const char *) download->title,
                        1, download->size > 0
                           ? 100 * download->bytes / download->size
                           : 0,
                        -1);
            }
    }
}

void Control::updateDownload (Node *node, int percentage)
{
    GtkTreeIter iter;
    GtkTreeModel *model;
    GtkListStore *store;

    if (!m_app->downloads_view && !percentage)
        showDownloadPage (NULL, m_app);

    if (m_app->downloads_view) {
        model = gtk_tree_view_get_model (GTK_TREE_VIEW (m_app->downloads_view));
        gtk_tree_model_get_iter_first (model, &iter);
        store = GTK_LIST_STORE (model);
    }
    for (Node *c = downloads ()->firstChild (); c; c = c->nextSibling ()) {
        if (c == node) {
            Download *download = static_cast <Download *> (c);
            if (m_app->downloads_view)
                gtk_list_store_set (store, &iter, 1, percentage, -1);
            if (100 == percentage)
                addRecent (download->file, download->title);
            break;
        }
        if (id_node_download == c->id && m_app->downloads_view)
            gtk_tree_model_iter_next (model, &iter);
    }
}

void Control::videoDetected ()
{
    Mrl *mrl = m_process ? m_process->mrl () : NULL;
    if (mrl && mrl->audio_only) {
        mrl->audio_only = false;
        viewArea ()->setAudioVideoNode (mrl);
        updateViewAreaSizes ();
    }
    app_full_screen (m_app);
}

static gboolean downloadPageDestroyed (GtkObject *, Application *app)
{
    debugLog() << "Download page destroyed\n";
    app->window_downloads = NULL;
    app->downloads_view = NULL;
}

namespace {

struct DownloadInfo {
    HildonTouchSelector *selector;
    int selected;
};

}

static void downloadResponse (GtkDialog *, gint response, gpointer data)
{
    DownloadInfo *info = (DownloadInfo *) data;
    if (response == GTK_RESPONSE_ACCEPT)
        info->selected = hildon_touch_selector_get_active (info->selector, 0);
    else
        info->selected = -1;
}

static void
downloadTapped (GtkTreeView *view, GtkTreePath *path, Application *app)
{
    Control *ctr = CONTROL(app);
    Download *download = NULL;
    NodePtr safe;
    int i = gtk_tree_path_get_indices (path) [gtk_tree_path_get_depth (path)-1];
    for (Node *c = ctr->downloads ()->firstChild (); c; c = c->nextSibling ())
        if (id_node_download == c->id && !i--) {
            download = static_cast <Download *> (c);
            safe = download;
            break;
        }
    if (download) {
        DownloadInfo info;
        GtkWidget *dialog = gtk_dialog_new_with_buttons (download->title,
                app->window_downloads,
                GtkDialogFlags(GTK_DIALOG_MODAL|GTK_DIALOG_DESTROY_WITH_PARENT),
                GTK_STOCK_OK,
                GTK_RESPONSE_ACCEPT,
                NULL);
        g_signal_connect (G_OBJECT (dialog), "response",
                G_CALLBACK (downloadResponse), &info);
        GtkWidget *content = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
        gtk_box_set_homogeneous (GTK_BOX (content), false);

        GtkWidget *widget = gtk_label_new ("");
        gtk_label_set_justify (GTK_LABEL (widget), GTK_JUSTIFY_LEFT);
        gtk_label_set_markup (GTK_LABEL (widget),
                String ("Link <small><tt>") + download->uri +
                String ("</tt></small> saved as <small><tt>") +
                String (strrchr (download->file, '/') + 1) +
                String ("</tt></small>"));
        gtk_label_set_line_wrap (GTK_LABEL (widget), true);
        gtk_label_set_selectable (GTK_LABEL (widget), true);
        gtk_box_pack_start (GTK_BOX (content), widget, false, true, 5);

        widget = hildon_touch_selector_new_text ();
        info.selector = HILDON_TOUCH_SELECTOR (widget);
        hildon_touch_selector_append_text (info.selector,
                download->job
                    ? "Suspend Download"
                    : download->getAttribute ("size").isEmpty ()
                        ? "Restart Download"
                        : "Resume Download");
        hildon_touch_selector_append_text (info.selector,
                "Delete but keep file on disk");
        hildon_touch_selector_append_text (info.selector,
                "Delete and delete file on disk");
        gtk_box_pack_start (GTK_BOX (content), widget, true, true, 10);

        gtk_widget_show_all (content);
        gtk_window_set_default_size (GTK_WINDOW (dialog), 800, 320);
        gtk_dialog_run (GTK_DIALOG (dialog));
        gtk_widget_destroy (dialog);

        NodePtrW guard = safe;
        safe = NULL;
        if (guard) {
            struct stat st;
            switch (info.selected) {
            case 0:
                if (download->job) {
                    download->job->setError ();
                    download->job->kill (false);
                } else {
                    download->begin ();
                }
                break;
            case 1:
                ctr->addRecent (download->file, download->title);
                download->parentNode ()->removeChild (download);
                break;
            case 2:
                if (!stat (download->file, &st) && S_ISREG (st.st_mode))
                    unlink (download->file);
                download->parentNode ()->removeChild (download);
                break;
            }
        }
        ctr->updateDownloads ();
    }
}

static void addPlainDownload (GtkButton *, Application *app)
{
    String url = getUserInput (app, "Download URL",
            "Please enter the link to download", "http://");
    if (!url.isEmpty () && url != "http://") {
        Control *ctrl = CONTROL(app);
        ctrl->downloader->createDownload (url, String ());
    }
}

void showDownloadPage (GtkMenuItem *, Application *app)
{
    Control *ctrl = CONTROL(app);
    if (!app->window_downloads) {
        HildonWindow *win = HILDON_WINDOW (hildon_stackable_window_new ());
        gtk_window_set_title (GTK_WINDOW (win), "KMPlayer Download Page");
        g_signal_connect (G_OBJECT (win), "destroy",
                GTK_SIGNAL_FUNC (downloadPageDestroyed), app);

        GtkListStore *store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_INT);
        app->downloads_view = hildon_gtk_tree_view_new_with_model (HILDON_UI_MODE_EDIT, GTK_TREE_MODEL (store));
        GtkWidget *scroller = hildon_pannable_area_new ();

        g_object_unref (store);

        GtkTreeViewColumn *col = gtk_tree_view_column_new();
        gtk_tree_view_column_set_alignment (col, 0.5);
        gtk_tree_view_column_set_sizing (col, GTK_TREE_VIEW_COLUMN_FIXED);
        gtk_tree_view_column_set_fixed_width (col, 800);
        GtkCellRenderer *render = gtk_cell_renderer_progress_new ();
        gtk_cell_renderer_set_fixed_size (render, 800, -1);
        gtk_tree_view_column_pack_start (col, render, false);
        gtk_tree_view_append_column(GTK_TREE_VIEW(app->downloads_view), col);
        gtk_tree_view_column_set_attributes(col, render,
                "text", 0,
                "value", 1,
                NULL);
        g_signal_connect (G_OBJECT (app->downloads_view),
                "hildon-row-tapped", G_CALLBACK (downloadTapped), app);

        gtk_container_add (GTK_CONTAINER (scroller), app->downloads_view);
        gtk_container_add (GTK_CONTAINER (win), scroller);
        gtk_widget_show_all (scroller);

        HildonAppMenu *menu = HILDON_APP_MENU (hildon_app_menu_new ());
        GtkWidget *item = hildon_gtk_button_new (HILDON_SIZE_AUTO);
        gtk_button_set_label (GTK_BUTTON (item), "Add plain Download");
        hildon_app_menu_append (menu, GTK_BUTTON (item));
        g_signal_connect_after (G_OBJECT (item), "clicked",
                G_CALLBACK (addPlainDownload), app);
        gtk_widget_show_all (GTK_WIDGET (menu));
        hildon_window_set_app_menu (HILDON_WINDOW (win), menu);

        app->window_downloads = GTK_WINDOW (win);
        hildon_program_add_window (HILDON_PROGRAM (app->program), win);
    }
    ctrl->updateDownloads ();
    gtk_tree_selection_unselect_all (gtk_tree_view_get_selection (
                GTK_TREE_VIEW(app->downloads_view)));
    gtk_widget_show_all (GTK_WIDGET (app->window_downloads));
    app_update_screen_blanking (app);
}

/*
 * TODO
static void cb_iap (struct iap_event_t * event, void * arg) {
    if (event->type == OSSO_IAP_CONNECTED) {
    }
}

osso_iap_cb (cb_iap);
osso_iap_connect (OSSO_IAP_ANY, OSSO_IAP_REQUESTED_CONNECT, 0L);

'dbus-send --print-reply --dest=com.nokia.osso_browser /usr/bin/browser com.nokia.osso_browser.load_url string:%s' % url
*/
